import numpy as np
import pandas as pd
import random

from KerasGA.Selection import Selection
from KerasGA.Crossover import Crossover
from KerasGA.Mutation import Mutation

class GeneticAlgorithm(object):

    def __init__(self, model, population_size = 500, selection_rate = 0.4, mutation_rate = 0.3, 
                selection='roulette_selection',crossover='uniform',mutation='random_mutation'):
        self.model = model
        self.population_size = population_size
        self.parents = int(population_size * selection_rate)
        self.mutation_rate = mutation_rate
        self.selection = Selection(selection)
        self.crossover = Crossover(crossover)
        self.mutation = Mutation(mutation,mutation_rate)

    def generate_weights(self,shape):
        weights = []
        for _ in range(shape[0]):
            node = []
            for _ in range(shape[1]):
                random_value = random.uniform(-1, 1)
                node.append(random_value)
            weights.append(np.array(node,dtype= 'float32'))
            
        weights = np.array(weights,dtype= 'float32')
        biases = np.zeros(shape[1],dtype= 'float32')
        return [weights,biases]

    def generate_chromosome(self):
        chromosome = []
        for layer in self.model.layers:
            for weights in layer.weights:
                if len(weights.shape) != 1 :
                    chromosome += self.generate_weights(weights.shape)
        return chromosome

    def initial_population(self):
        chromosomes = []
        for _ in range(self.population_size):
            chromosomes.append(self.generate_chromosome())
        return chromosomes
    
    def strongest_parents(self, population, scores):
        scores_for_chromosomes = []
        for i in range(0, len(population)):
            scores_for_chromosomes.append(( population[i],scores[i] ))
        scores_for_chromosomes.sort(key=lambda x: x[1])
        top_performers = scores_for_chromosomes[-self.parents:]
        return top_performers
    
    def pair(self,strongest_parents):
        total_parents_score = sum([x[1] for x in strongest_parents])
        pick = random.uniform(0, total_parents_score)
        return [self.selection(strongest_parents, pick), self.selection(strongest_parents, pick)]


import numpy as np
import copy
import random

mutation_rate = 0,3

def Mutation(algorithm,mutation_r):
    global mutation_rate
    mutation_rate = mutation_r
    return random_mutation

def random_mutation(base_offsprings):
    offsprings = []
    for offspring in base_offsprings:
        offspring_mutation = copy.deepcopy(offspring)
        for layer in range(len(offspring)):
            for i in range(offspring[layer].shape[0]):
                if offspring[layer].ndim != 1 :
                    for j in range(offspring[layer].shape[1]):
                        if np.random.choice([True, False], p=[mutation_rate, 1-mutation_rate]):
                            offspring_mutation[layer][i][j] = random.uniform(-1, 1)
        offsprings.append(offspring_mutation)
    return offsprings


from random import shuffle

def Selection(algorithm):
    return roulette_selection

def roulette_selection(parents, pick):
    shuffle(parents)
    current = 0
    for parent in parents:
        current += parent[1]
        if current >= pick:
            return parent


import random

def Crossover(algorithm):
    return uniform

def uniform(x, y):
    offspring_x = x
    offspring_y = y
    for layer in range(len(x)):
        for i in range(x[layer].shape[0]):
            if x[layer].ndim != 1 :
                for j in range(x[layer].shape[1]):
                    if random.choice([True, False]):
                        offspring_x[layer][i][j] = y[layer][i][j]
                        offspring_y[layer][i][j] = x[layer][i][j]
    return offspring_x, offspring_y